#include "RR_Sparki.h"
#include "RR_Accelerometer.h"

extern RR_SparkiClass sparki;
extern float xAxisAccel;
extern float yAxisAccel;
extern float zAxisAccel;
extern char st7565_buffer[1024];
extern unsigned char sweep_radar[100];
extern int sweep_position;
extern int sweep_enabled;
extern uint8_t xUpdateMin, xUpdateMax, yUpdateMin, yUpdateMax;

RR_Accelerometer accel;

unsigned int g_crc;
unsigned int g_command;
unsigned int g_value;
unsigned int g_valueLow;
unsigned int g_valueHigh;
int g_value1;
int g_value2;
int g_value3;
unsigned long g_heartBeat=0;
unsigned char g_store[65];
unsigned char g_photoState = 0;
int g_lastIRPulse=0;
int g_lastIRCode=255;
int g_stringIndex=0;
unsigned char serial_mode = 0;

float noteToFreq[] PROGMEM = {16.35f, 17.32f, 18.35f, 19.45f, 20.60f, 21.83f, 23.12f, 24.50f, 25.96f, 27.50f, 29.14f, 30.87f, 32.70f, 34.65f, 36.71f, 38.89f, 41.20f, 43.65f, 46.25f, 49.00f, 51.91f, 55.00f, 58.27f, 61.74f, 65.41f, 69.30f, 73.42f, 77.78f, 82.41f, 87.31f, 92.50f, 98.00f, 103.83f, 110.00f, 116.54f, 123.47f, 130.81f, 138.59f, 146.83f, 155.56f, 164.81f, 174.61f, 185.00f, 196.00f, 207.65f, 220.00f, 233.08f, 246.94f, 261.63f, 277.18f, 293.66f, 311.13f, 329.63f, 349.23f, 369.99f, 392.00f, 415.30f, 440.00f, 466.16f, 493.88f, 523.25f, 554.37f, 587.33f, 622.25f, 659.26f, 698.46f, 739.99f, 783.99f, 830.61f, 880.00f, 932.33f, 987.77f, 1046.50f, 1108.73f, 1174.66f, 1244.51f, 1318.51f, 1396.91f, 1479.98f, 1567.98f, 1661.22f, 1760.00f, 1864.66f, 1975.53f, 2093.00f, 2217.46f, 2349.32f, 2489.02f, 2637.02f, 2793.83f, 2959.96f, 3135.96f, 3322.44f, 3520.00f, 3729.31f, 3951.07f, 4186.01f, 4434.92f, 4698.64f, 4978.03f};

#define SPARKI_GET_ID 0
#define SPARKI_SET_MOTORS 1
#define SPARKI_SET_LED 2
#define SPARKI_SET_GRIPPER 3
#define SPARKI_PLAY 4
#define SPARKI_GET_PHOTO 5
#define SPARKI_GET_LINE 6
#define SPARKI_SET_LCD_TEXT 7
#define SPARKI_SET_LCD_IMAGE 8
#define SPARKI_GET_ACCELEROMETER 9
#define SPARKI_GET_REMOTE 10
#define SPARKI_SET_REMOTE 11
#define SPARKI_GET_RADAR 12
#define SPARKI_STOP_RADAR 13
#define SPARKI_GET_COMPASS 14

#define SERVO               9        //

void setup() 
{
  // prevent the head from turning quickly on startup
  sparki.begin();  
  sparki.startServoTimer();
  accel.begin();
  Serial.begin(115200);
  Serial1.begin(9600);
}

void readPacket()
{
  // get header byte
  // 128 (bit 8) flag indicates a new g_command packet .. that 
  // means the g_value bytes can never have 128 set!
  // next byte is the g_command 0-8
  // next byte is the g_channel 0-16
  do
  {
    g_command = Serial_read();
  }
  while ((g_command&128)==0);
  
  g_command^=128;
}

int readValuePacket()
{
  unsigned int g_valueLow;
  unsigned int g_valueHigh;
  
  // wait for g_value low byte    
  g_valueLow = Serial_read();
  if (g_valueLow&128) return 0;

  // wait for g_value high byte    
  g_valueHigh = Serial_read();
  if (g_valueHigh&128) return 0;
    
  // wait for g_crc byte    
  g_crc = Serial_read();
  if (g_crc&128) return 0;
  
  if (g_crc!=(((128|g_command)^g_valueLow^g_valueHigh)&127)) return 0;

  return 1;
}

int readDoubleValuePacket()
{
  unsigned int low1;
  unsigned int high1;
  unsigned int low2;
  unsigned int high2;
  
  // wait for value1 low byte    
  low1 = Serial_read();
  if (low1&128) return 0;

  // wait for value1 high byte    
  high1 = Serial_read();
  if (high1&128) return 0;
    
  // wait for value2 low byte    
  low2 = Serial_read();
  if (low2&128) return 0;

  // wait for value2 high byte    
  high2 = Serial_read();
  if (high2&128) return 0;
    
  // wait for g_crc byte    
  g_crc = Serial_read();
  if (g_crc&128) return 0;
  
  if (g_crc!=(((128|g_command)^low1^high1^low2^high2)&127)) return 0;

  // create 15 bit value with sign extended
  g_value1 = low1|(high1<<7)|((high1&64)<<9);
  g_value2 = low2|(high2<<7)|((high2&64)<<9);
  
  return 1;
}

int readTripleValuePacket()
{
  unsigned int low1;
  unsigned int high1;
  unsigned int low2;
  unsigned int high2;
  unsigned int low3;
  unsigned int high3;
  
  // wait for value1 low byte    
  low1 = Serial_read();
  if (low1&128) return 0;

  // wait for value1 high byte    
  high1 = Serial_read();
  if (high1&128) return 0;
    
  // wait for value2 low byte    
  low2 = Serial_read();
  if (low2&128) return 0;

  // wait for value2 high byte    
  high2 = Serial_read();
  if (high2&128) return 0;
    
  // wait for value2 low byte    
  low3 = Serial_read();
  if (low3&128) return 0;

  // wait for value2 high byte    
  high3 = Serial_read();
  if (high3&128) return 0;
    
  // wait for g_crc byte    
  g_crc = Serial_read();
  if (g_crc&128) return 0;
  
  if (g_crc!=(((128|g_command)^low1^high1^low2^high2^low3^high3)&127)) return 0;

  // create 15 bit value with sign extended
  g_value1 = low1|(high1<<7)|((high1&64)<<9);
  g_value2 = low2|(high2<<7)|((high2&64)<<9);
  g_value3 = low3|(high3<<7)|((high3&64)<<9);
  
  return 1;
}

int Serial_available()
{
  if (serial_mode==1)
    return Serial.available();
  else
    return Serial1.available();
}

int Serial_print(char *txt)
{
  if (serial_mode==1)
    return Serial.print(txt);
  else
    return Serial1.print(txt);
}

// if we don't see a character in about 1 second ... bail
int Serial_read()
{
  unsigned long int start = micros();
  
  if (serial_mode==1)
  {
    while (Serial.available() <= 0) { if ((micros() - start)>1000000) return -1; }
    return Serial.read();
  }
  else
  {
    while (Serial1.available() <= 0) { if ((micros() - start)>1000000) return -1; }
    return Serial1.read();
  }
}

int Serial_write(unsigned char *buffer, int len)
{
  if (serial_mode==1)
    return Serial.write(buffer, len);
  else
    return Serial1.write(buffer, len);
}

int Serial_readBytes(char *txt, int len)
{
  if (serial_mode==1)
    return Serial.readBytes(txt, len);
  else
    return Serial1.readBytes(txt, len);
}

void readAccelData(float *arr)
{
  sparki.readAccelData();
  arr[0] = xAxisAccel;
  arr[1] = yAxisAccel;
  arr[2] = zAxisAccel;
}

void loop() 
{
  int i;
  int bank;
  int len;

  // stop the sparki if we don't hear anything from
  // the PC in 2 seconds
  if ((micros() - g_heartBeat)>2000000)
  {
    sparki.moveStop();
    serial_mode=0;
  }
  
  if (serial_mode==0)
  {
    // determine if we are connected via cable or bluetooth
    if (Serial.available()>0)
      serial_mode=1;
    else
    if (Serial1.available()>0)
      serial_mode=2;
    else 
      return;
  }
  
  while (Serial_available()>0)
  {
    g_heartBeat = micros();

    readPacket();
    
    switch (g_command)
    {
      // init
      case  SPARKI_GET_ID:
        Serial_print("SPARKI1");
      break;
      case  SPARKI_SET_MOTORS:
        if (readDoubleValuePacket())
        {
          if (g_value1>0)
            sparki.motorRotate(MOTOR_LEFT, DIR_CCW, g_value1);
          else  
          if (g_value1<0)
            sparki.motorRotate(MOTOR_LEFT, DIR_CW, -g_value1);
          else
            sparki.motorStop(MOTOR_LEFT);
          
          if (g_value2>0)
            sparki.motorRotate(MOTOR_RIGHT, DIR_CW, g_value2);
          else  
          if (g_value2<0)
            sparki.motorRotate(MOTOR_RIGHT, DIR_CCW, -g_value2);
          else
            sparki.motorStop(MOTOR_RIGHT);
        }
      break;
      case  SPARKI_SET_LED:
        if (readTripleValuePacket())
        {
          sparki.RGB(g_value1, g_value2, g_value3);
        }
      break;
      case  SPARKI_SET_GRIPPER:
        g_value = Serial_read();
        switch (g_value)
        {
         case 0: sparki.gripperStop(); break;
         case 2: sparki.gripperCloseSteps(); break;
         case 1: sparki.gripperOpenSteps(); break;
        } 
      break;
      case  SPARKI_PLAY:
        g_value = Serial_read();
        Serial_readBytes((char *)g_store, g_value);
        for (i=0;i<g_value;i+=2)
        {
          g_value1 = g_store[i];
          g_value2 = ((int)g_store[i+1])*15;
          tone(BUZZER, pgm_read_float(noteToFreq+g_value1), g_value2);
          delay(g_value2 + (g_value2>>2));
          noTone(BUZZER);
        }
      break;
      case  SPARKI_GET_PHOTO:
      {
        int left   = sparki.lightLeft();   // measure the left light sensor
        int center = sparki.lightCenter(); // measure the center light sensor
        int right  = sparki.lightRight();  // measure the right light sensor    
    
        g_store[0]=left;
        g_store[1]=left>>8;
        g_store[2]=center;
        g_store[3]=center>>8;
        g_store[4]=right;
        g_store[5]=right>>8;
      
        Serial_write(g_store, 6);
      }
      break;
      case  SPARKI_GET_LINE:
      {
        int edgeLeft   = sparki.edgeLeft();
        int lineLeft   = sparki.lineLeft();
        int lineCenter = sparki.lineCenter();
        int lineRight  = sparki.lineRight();
        int edgeRight  = sparki.edgeRight();
    
        g_store[0]=edgeLeft&255;
        g_store[1]=edgeLeft>>8;
        g_store[2]=lineLeft&255;
        g_store[3]=lineLeft>>8;
        g_store[4]=lineCenter&255;
        g_store[5]=lineCenter>>8;
        g_store[6]=lineRight&255;
        g_store[7]=lineRight>>8;
        g_store[8]=edgeRight&255;
        g_store[9]=edgeRight>>8;
      
        Serial_write(g_store, 10);
      }
      break;
      case SPARKI_SET_LCD_TEXT:
        // read in string length, will be in blocks of 64 since reading in longer
        // strings causes irregular behaviour
        len = Serial_read();
        if (len==0)
        {
          g_store[g_stringIndex]=0;
          sparki.clearLCD();
          sparki.drawString(0, 0, (char *)g_store);     
          sparki.updateLCD();       
          g_stringIndex=0;
        }
        else
        {
          Serial_readBytes((char *)&g_store[g_stringIndex], len);
          g_stringIndex+=len;
        }
      break;
      case SPARKI_SET_LCD_IMAGE:
        bank = Serial_read();
        Serial_readBytes(&st7565_buffer[bank*64], 64);
        // when we receive the last bank, update the image
        if (bank==15)
          sparki.updateAll();
      break;
      case SPARKI_GET_ACCELEROMETER:
        //readAccelData((float *)g_store);
        accel.readGs((float *)&g_store);
        Serial_write(g_store, 12);
      break;
      case SPARKI_GET_REMOTE:
        g_store[0] = sparki.readIR();
        // eliminate fast code pulses
        if (g_store[0]==255)
        {
          if (g_lastIRPulse<20)
            g_store[0] = g_lastIRCode;
           
          g_lastIRPulse++; 
        }
        else
        {
          g_lastIRPulse=0;
          g_lastIRCode = g_store[0];
        }
        Serial_write(g_store, 1);
      break;
      case SPARKI_SET_REMOTE:
        g_value1 = Serial_read();
        //sparki.sendIR(g_value1);
      break;
      case SPARKI_GET_RADAR:
        sweep_enabled=1;
        pinMode(SERVO,OUTPUT);
        Serial_write((unsigned char *)&sweep_position, 2);
        Serial_write((unsigned char *)sweep_radar, 100);
      break;
      case SPARKI_STOP_RADAR:
        sweep_enabled=0;
        pinMode(SERVO,INPUT);
      break;
      case SPARKI_GET_COMPASS:
        sparki.getRawCompass(g_store);
        Serial_write((unsigned char *)g_store, 4);
      break;
    }
  }
}  



